from uiwiz.element import Element


class Container(Element):
    root_class: str = "container flex flex-col mx-auto {max_w} {padding} {space_y} grow"

    def __init__(self, max_w: str = "max-w-[960px]", padding: str = "pt-4 pb-4", space_y: str = "space-y-4") -> None:
        """Container

        A container element that centers a box in the middle of the screen.
        """
        super().__init__()
        self.__root_class__ = Container.root_class.format(max_w=max_w, padding=padding, space_y=space_y)
        self.classes()



from contextlib import asynccontextmanager

import uvicorn
from fastapi import Request
from fastapi.responses import HTMLResponse

from docs.layout import Layout, Page, pages
from docs.page_docs import docs_router
from uiwiz import PageDefinition, PageRouter, UiwizApp, ui
from uiwiz.frame import Frame

router: PageRouter = PageRouter(page_definition_class=PageDefinition)
page_dict: dict[str, Page] = {item.path: item for item in pages}


@asynccontextmanager
async def lifespan(app: UiwizApp):
    """Lifespan event handler for the application."""

    for page in pages:
        """
        Register each page with the application.
        This is the same as using the @app.page decorator,
        but allows for dynamic page registration.
        """
        if page.content is None:
            continue
        app.page(path=page.path, title=page.title)(render_md)
        app.include_router(docs_router)

    yield


app: UiwizApp = UiwizApp(lifespan=lifespan, page_definition_class=Layout)


async def render_md(request: Request):
    page = page_dict.get(request.url.path, None)
    if page:
        await page.render()
    else:
        with ui.container(padding="p-4"):
            ui.markdown("Page not found.")


async def not_found():
    with ui.container(padding="p-4"):
        ui.markdown("Page not found. Please check the URL or return to the home page.")
        for page in pages:
            ui.link(page.title, page.path)


@app.exception_handler(404)
async def not_found_exception_handler(request: Request, exc: Exception):
    await app.page_definition_class().render(not_found, request, title="Not Found")
    return HTMLResponse(
        content=Frame.get_stack().render(),
        status_code=404,
        media_type="text/html",
    )


if __name__ == "__main__":
    uvicorn.run("docs.main:app", host="0.0.0.0", port=8080, reload=True)



from typing import Optional

from uiwiz.element import Element


class Tabs(Element):
    root_class: str = "tabs "
    _classes: str = "tabs-box"

    def __init__(self) -> None:
        """Tabs
        This element is used for tab navigation.
        It should be used as a context manager to create tabs.

        .. code-block:: python
            from uiwiz import ui

            with ui.tabs():
                with ui.tab("Tab 1").active():
                    ui.label("Content of Tab 1")
                with ui.tab("Tab 2"):
                    ui.label("Content of Tab 2")
                with ui.tab("Tab 3"):
                    ui.label("Content of Tab 3")

        """
        super().__init__()
        self.classes(Tabs.root_class + Tabs._classes)
        self.attributes["role"] = "tablist"

    def __exit__(self, *_):
        super().__exit__(*_)
        has_active = False
        first_child_tab = None
        for child in self.children:
            if not isinstance(child, Tab):
                continue

            if first_child_tab is None:
                first_child_tab = child

            if "checked" in child.attributes:
                has_active = True
        if not has_active:
            first_child_tab.active()


class Tab(Element):
    _classes: str = "tab"

    def __init__(self, title: str, active: Optional[bool] = None) -> None:
        """Tab

        This element is used for tab navigation
        and should be used inside a :class:`Tabs` element.
        The first tab will be active by default if no active tab is set.

        .. code-block:: python
            from uiwiz import ui

            with ui.tabs():
                with ui.tab("Tab 1").active():
                    ui.label("Content of Tab 1")
                with ui.tab("Tab 2"):
                    ui.label("Content of Tab 2")
                with ui.tab("Tab 3"):
                    ui.label("Content of Tab 3")

        :param title: The title of the tab
        :param active: If the tab should be active by default. Defaults to None, which will make the first tab active.
        """
        self.selector = Element("input")
        self.selector.attributes["name"] = self.selector.parent_element.id
        self.selector.attributes["type"] = "radio"
        self.selector.attributes["aria-label"] = title
        self.selector.classes(self._classes)
        if active:
            self.active()

        super().__init__()
        self.classes("tab-content bg-base-100 border-base-300 p-4")
        self.attributes["role"] = "tabpanel"

    def active(self) -> "Tab":
        self.selector.attributes["checked"] = ""
        return self



from uiwiz.element import Element


class Footer(Element):
    root_class: str = "mt-auto "
    _classes: str = "items-center p-4 bg-neutral text-neutral-content"

    def __init__(self) -> None:
        """Footer

        A footer is a section at the bottom of a page that contains information about the page.

        .. code-block:: python
            from uiwiz import ui

            with ui.footer():
                ui.label("Footer content")
        """
        super().__init__(tag="footer")
        self.classes(Footer._classes)



WONDERLAND = {
    "version": 1,
    "themeName": "wonderland",
    "theme": {
        "seriesCnt": "3",
        "backgroundColor": "rgba(255,255,255,0)",
        "titleColor": "#666666",
        "subtitleColor": "#999999",
        "textColorShow": False,
        "textColor": "#333",
        "markTextColor": "#ffffff",
        "color": ["#4ea397", "#22c3aa", "#7bd9a5", "#d0648a", "#f58db2", "#f2b3c9"],
        "borderColor": "#ccc",
        "borderWidth": 0,
        "visualMapColor": ["#d0648a", "#22c3aa", "#adfff1"],
        "legendTextColor": "#999999",
        "kColor": "#d0648a",
        "kColor0": "transparent",
        "kBorderColor": "#d0648a",
        "kBorderColor0": "#22c3aa",
        "kBorderWidth": "1",
        "lineWidth": "3",
        "symbolSize": "8",
        "symbol": "emptyCircle",
        "symbolBorderWidth": "2",
        "lineSmooth": 1,
        "graphLineWidth": "1",
        "graphLineColor": "#cccccc",
        "mapLabelColor": "#28544e",
        "mapLabelColorE": "#349e8e",
        "mapBorderColor": "#999999",
        "mapBorderColorE": "#22c3aa",
        "mapBorderWidth": 0.5,
        "mapBorderWidthE": 1,
        "mapAreaColor": "#eeeeee",
        "mapAreaColorE": "rgba(34,195,170,0.25)",
        "axes": [
            {
                "type": "all",
                "name": "通用坐标轴",
                "axisLineShow": True,
                "axisLineColor": "#cccccc",
                "axisTickShow": False,
                "axisTickColor": "#333",
                "axisLabelShow": True,
                "axisLabelColor": "#999999",
                "splitLineShow": True,
                "splitLineColor": ["#eeeeee"],
                "splitAreaShow": False,
                "splitAreaColor": ["rgba(250,250,250,0.05)", "rgba(200,200,200,0.02)"],
            },
            {
                "type": "category",
                "name": "类目坐标轴",
                "axisLineShow": True,
                "axisLineColor": "#333",
                "axisTickShow": True,
                "axisTickColor": "#333",
                "axisLabelShow": True,
                "axisLabelColor": "#333",
                "splitLineShow": False,
                "splitLineColor": ["#ccc"],
                "splitAreaShow": False,
                "splitAreaColor": ["rgba(250,250,250,0.3)", "rgba(200,200,200,0.3)"],
            },
            {
                "type": "value",
                "name": "数值坐标轴",
                "axisLineShow": True,
                "axisLineColor": "#333",
                "axisTickShow": True,
                "axisTickColor": "#333",
                "axisLabelShow": True,
                "axisLabelColor": "#333",
                "splitLineShow": True,
                "splitLineColor": ["#ccc"],
                "splitAreaShow": False,
                "splitAreaColor": ["rgba(250,250,250,0.3)", "rgba(200,200,200,0.3)"],
            },
            {
                "type": "log",
                "name": "对数坐标轴",
                "axisLineShow": True,
                "axisLineColor": "#333",
                "axisTickShow": True,
                "axisTickColor": "#333",
                "axisLabelShow": True,
                "axisLabelColor": "#333",
                "splitLineShow": True,
                "splitLineColor": ["#ccc"],
                "splitAreaShow": False,
                "splitAreaColor": ["rgba(250,250,250,0.3)", "rgba(200,200,200,0.3)"],
            },
            {
                "type": "time",
                "name": "时间坐标轴",
                "axisLineShow": True,
                "axisLineColor": "#333",
                "axisTickShow": True,
                "axisTickColor": "#333",
                "axisLabelShow": True,
                "axisLabelColor": "#333",
                "splitLineShow": True,
                "splitLineColor": ["#ccc"],
                "splitAreaShow": False,
                "splitAreaColor": ["rgba(250,250,250,0.3)", "rgba(200,200,200,0.3)"],
            },
        ],
        "axisSeperateSetting": False,
        "toolboxColor": "#999999",
        "toolboxEmphasisColor": "#666666",
        "tooltipAxisColor": "#cccccc",
        "tooltipAxisWidth": 1,
        "timelineLineColor": "#4ea397",
        "timelineLineWidth": 1,
        "timelineItemColor": "#4ea397",
        "timelineItemColorE": "#4ea397",
        "timelineCheckColor": "#4ea397",
        "timelineCheckBorderColor": "#3cebd2",
        "timelineItemBorderWidth": 1,
        "timelineControlColor": "#4ea397",
        "timelineControlBorderColor": "#4ea397",
        "timelineControlBorderWidth": 0.5,
        "timelineLabelColor": "#4ea397",
        "datazoomBackgroundColor": "rgba(255,255,255,0)",
        "datazoomDataColor": "rgba(222,222,222,1)",
        "datazoomFillColor": "rgba(114,230,212,0.25)",
        "datazoomHandleColor": "#cccccc",
        "datazoomHandleWidth": "100",
        "datazoomLabelColor": "#999999",
    },
}

WESTEROS = {
    "version": 1,
    "themeName": "westeros",
    "theme": {
        "seriesCnt": "4",
        "backgroundColor": "rgba(0,0,0,0)",
        "titleColor": "#516b91",
        "subtitleColor": "#93b7e3",
        "textColorShow": False,
        "textColor": "#333",
        "markTextColor": "#eeeeee",
        "color": ["#516b91", "#59c4e6", "#edafda", "#93b7e3", "#a5e7f0", "#cbb0e3"],
        "borderColor": "#ccc",
        "borderWidth": 0,
        "visualMapColor": ["#516b91", "#59c4e6", "#a5e7f0"],
        "legendTextColor": "#999999",
        "kColor": "#edafda",
        "kColor0": "transparent",
        "kBorderColor": "#d680bc",
        "kBorderColor0": "#8fd3e8",
        "kBorderWidth": "2",
        "lineWidth": "2",
        "symbolSize": "6",
        "symbol": "emptyCircle",
        "symbolBorderWidth": "2",
        "lineSmooth": True,
        "graphLineWidth": 1,
        "graphLineColor": "#aaaaaa",
        "mapLabelColor": "#000",
        "mapLabelColorE": "#516b91",
        "mapBorderColor": "#516b91",
        "mapBorderColorE": "#516b91",
        "mapBorderWidth": 0.5,
        "mapBorderWidthE": 1,
        "mapAreaColor": "#f3f3f3",
        "mapAreaColorE": "#a5e7f0",
        "axes": [
            {
                "type": "all",
                "name": "通用坐标轴",
                "axisLineShow": True,
                "axisLineColor": "#cccccc",
                "axisTickShow": False,
                "axisTickColor": "#333",
                "axisLabelShow": True,
                "axisLabelColor": "#999999",
                "splitLineShow": True,
                "splitLineColor": ["#eeeeee"],
                "splitAreaShow": False,
                "splitAreaColor": ["rgba(250,250,250,0.05)", "rgba(200,200,200,0.02)"],
            },
            {
                "type": "category",
                "name": "类目坐标轴",
                "axisLineShow": True,
                "axisLineColor": "#333",
                "axisTickShow": True,
                "axisTickColor": "#333",
                "axisLabelShow": True,
                "axisLabelColor": "#333",
                "splitLineShow": False,
                "splitLineColor": ["#ccc"],
                "splitAreaShow": False,
                "splitAreaColor": ["rgba(250,250,250,0.3)", "rgba(200,200,200,0.3)"],
            },
            {
                "type": "value",
                "name": "数值坐标轴",
                "axisLineShow": True,
                "axisLineColor": "#333",
                "axisTickShow": True,
                "axisTickColor": "#333",
                "axisLabelShow": True,
                "axisLabelColor": "#333",
                "splitLineShow": True,
                "splitLineColor": ["#ccc"],
                "splitAreaShow": False,
                "splitAreaColor": ["rgba(250,250,250,0.3)", "rgba(200,200,200,0.3)"],
            },
            {
                "type": "log",
                "name": "对数坐标轴",
                "axisLineShow": True,
                "axisLineColor": "#333",
                "axisTickShow": True,
                "axisTickColor": "#333",
                "axisLabelShow": True,
                "axisLabelColor": "#333",
                "splitLineShow": True,
                "splitLineColor": ["#ccc"],
                "splitAreaShow": False,
                "splitAreaColor": ["rgba(250,250,250,0.3)", "rgba(200,200,200,0.3)"],
            },
            {
                "type": "time",
                "name": "时间坐标轴",
                "axisLineShow": True,
                "axisLineColor": "#333",
                "axisTickShow": True,
                "axisTickColor": "#333",
                "axisLabelShow": True,
                "axisLabelColor": "#333",
                "splitLineShow": True,
                "splitLineColor": ["#ccc"],
                "splitAreaShow": False,
                "splitAreaColor": ["rgba(250,250,250,0.3)", "rgba(200,200,200,0.3)"],
            },
        ],
        "axisSeperateSetting": False,
        "toolboxColor": "#999999",
        "toolboxEmphasisColor": "#666666",
        "tooltipAxisColor": "#cccccc",
        "tooltipAxisWidth": 1,
        "timelineLineColor": "#8fd3e8",
        "timelineLineWidth": 1,
        "timelineItemColor": "#8fd3e8",
        "timelineItemColorE": "#8fd3e8",
        "timelineCheckColor": "#8fd3e8",
        "timelineCheckBorderColor": "#8a7ca8",
        "timelineItemBorderWidth": 1,
        "timelineControlColor": "#8fd3e8",
        "timelineControlBorderColor": "#8fd3e8",
        "timelineControlBorderWidth": 0.5,
        "timelineLabelColor": "#8fd3e8",
        "datazoomBackgroundColor": "rgba(0,0,0,0)",
        "datazoomDataColor": "rgba(255,255,255,0.3)",
        "datazoomFillColor": "rgba(167,183,204,0.4)",
        "datazoomHandleColor": "#a7b7cc",
        "datazoomHandleWidth": "100",
        "datazoomLabelColor": "#333",
    },
}



from uiwiz import ui


def test_frame_id_reuse():
    # setup
    from uiwiz.frame import get_request

    get_request().headers = {"hx-target": "a-1", "hx-swap": "outerHTML"}

    # Init
    first_ele = ui.element()
    ele = ui.element()

    assert "a-1" == first_ele.id
    assert "a-1" != ele.id



from unittest import mock

from uiwiz import ui
from uiwiz.frame import Frame


def test_element_html():
    output = str(ui.element())
    assert '<div id="a-0"></div>' == output


def test_element_html_nested_random():
    output = ""
    with ui.element() as root:
        ui.element()
        output = str(root)
    assert '<div id="a-0"><div id="a-1"></div></div>' == output


def test_element_void():
    output = str(ui.element("img"))
    assert '<img id="a-0">' == output


def test_get_target_none():
    el = ui.element()
    assert "this" == el.__get_target__(None)


def test_get_target_lambda():
    el = ui.element()
    assert "#this" == el.__get_target__(lambda: "this")


def test_get_target_str():
    el = ui.element()
    assert "#outerHTML" == el.__get_target__("outerHTML")
    assert "#asd" == el.__get_target__("asd")
    assert "this" == el.__get_target__("this")
    assert "#shouldbeok" == el.__get_target__("#shouldbeok")


def test_get_target_element_id():
    el = ui.element()
    el2 = ui.element()
    assert f"#{el2.id}" == el.__get_target__(el2)


def test_oob_top_level():
    with ui.element() as root:
        ui.toast("uiwiz")
    assert '<div id="a-0"></div>' == str(root)


def test_event():
    func = lambda: ui.toast("test")
    html = str(ui.button("Click me").on_click(func))

    assert (
        f'<button id="a-0" class="btn btn-md" hx-target="this" hx-swap="outerHTML" hx-post="/_uiwiz/hash/{func.__hash__()}" hx-trigger="click" hx-ext="json-enc">Click me</button>'
        == html
    )


def test_before_render_lifecycle():
    el = ui.element()
    mo = mock.MagicMock()
    el.before_render = mo

    str(el)

    mo.assert_called_once()


def test_element_get_name():
    el = ui.element()
    assert el.name is None


def test_element_get_value():
    el = ui.element()
    assert el.value is None


def test_element_set_value():
    el = ui.element()
    el.value = "test"
    assert el.value == "test"


def test_element_get_classes():
    el = ui.element().classes("glass")
    assert el.get_classes() == "glass"


def test_add_script():
    script = 'console.log("test")'
    el = ui.element()
    el.script = script
    str(el)
    assert Frame.get_stack().scripts[0] == script


def test_oob_render_outside_of_parent():
    with ui.element():
        ui.element(oob=True)
    out = Frame.get_stack().render()
    assert out == '<div id="a-0"></div><div id="a-1" hx-swap-oob="true"></div>'


def test_str_html_escape():
    el = ui.element()
    el.attributes["value"] = "<script>alert('test')</script>"
    assert str(el) == '<div id="a-0" value="&lt;script&gt;alert(&#x27;test&#x27;)&lt;/script&gt;"></div>'


def test_bool_html_escape():
    el = ui.element()
    el.attributes["value"] = True
    assert str(el) == '<div id="a-0" value="true"></div>'


def test_number_html_escape():
    el = ui.element()
    el.attributes["value"] = 123
    assert str(el) == '<div id="a-0" value="123"></div>'


def test_none_html_escape():
    el = ui.element()
    el.attributes["value"] = None
    assert str(el) == '<div id="a-0" value="None"></div>'



import functools
import inspect
from functools import partial
from typing import Annotated, Any, Callable, Optional, Sequence, Type, TypedDict

from fastapi import APIRouter, Request, Response, params
from fastapi.datastructures import Default
from fastapi.responses import HTMLResponse, JSONResponse
from fastapi.routing import APIRoute
from starlette.types import ASGIApp, Lifespan
from typing_extensions import Doc

from uiwiz.element import Element
from uiwiz.frame import Frame
from uiwiz.page_definition import PageDefinition


class DecKwargs(TypedDict):
    page: PageDefinition
    request: Request
    response: Response


class PageRouter(APIRouter):
    """
    `PageRouter` class, used to group *path operations*, for example to structure
    an app in multiple files. It would then be included in the `UiWizard` app, or
    in another `PageRouter` (ultimately included in the app).

    ## Example

    ```python
    from uiwiz import PageRouter, UiwizApp, ui

    app = UiwizApp()
    router = PageRouter()

    @router.get("/users/", tags=["users"])
    async def read_users():
        ui.element("h1", content="Hello world")

    app.include_router(router)
    ```
    """

    def __init__(
        self,
        *,
        prefix: Annotated[str, Doc("An optional path prefix for the router.")] = "",
        dependencies: Annotated[
            Optional[Sequence[params.Depends]],
            Doc(
                """
                A list of dependencies (using `Depends()`) to be applied to all the
                *path operations* in this router.

                Read more about it in the
                [FastAPI docs for Bigger Applications - Multiple Files](https://fastapi.tiangolo.com/tutorial/bigger-applications/#include-an-apirouter-with-a-custom-prefix-tags-responses-and-dependencies).
                """
            ),
        ] = None,
        default_response_class: Annotated[
            Type[Response],
            Doc(
                """
                The default response class to be used.

                Read more in the
                [FastAPI docs for Custom Response - HTML, Stream, File, others](https://fastapi.tiangolo.com/advanced/custom-response/#default-response-class).
                """
            ),
        ] = Default(JSONResponse),
        redirect_slashes: Annotated[
            bool,
            Doc(
                """
                Whether to detect and redirect slashes in URLs when the client doesn't
                use the same format.
                """
            ),
        ] = True,
        default: Annotated[
            Optional[ASGIApp],
            Doc(
                """
                Default function handler for this router. Used to handle
                404 Not Found errors.
                """
            ),
        ] = None,
        route_class: Annotated[
            Type[APIRoute],
            Doc(
                """
                Custom route (*path operation*) class to be used by this router.

                Read more about it in the
                [FastAPI docs for Custom Request and APIRoute class](https://fastapi.tiangolo.com/how-to/custom-request-and-route/#custom-apiroute-class-in-a-router).
                """
            ),
        ] = APIRoute,
        # which the router cannot know statically, so we use typing.Any
        lifespan: Annotated[
            Optional[Lifespan[Any]],
            Doc(
                """
                A `Lifespan` context manager handler. This replaces `startup` and
                `shutdown` functions with a single context manager.

                Read more in the
                [FastAPI docs for `lifespan`](https://fastapi.tiangolo.com/advanced/events/).
                """
            ),
        ] = None,
        page_definition_class: Annotated[
            Optional[Type[PageDefinition]],
            Doc("""
                The page definition class to use for this router.
                
                This enables the use of custom page definitions for rendering the HTML
                pages. The default is `PageDefinition`, which provides a basic HTML
                structure. You can create your own class that inherits from `PageDefinition`
                and override the `header`, `body`, and `content` methods to customize the
                HTML structure and content as needed. Setting the `page_definition_class` in the 
                UiwizApp will set the default for all routers.
                Example:
                ```python
                class MyPageDefinition(PageDefinition):
                    def header(self, header: Element) -> None:
                        # Custom header content
                        Element("link", href="/custom.css", rel="stylesheet")
                
                    def body(self, body: Element) -> None:
                        # Custom body content
                        Element("div", content="Custom Body").classes("custom-body")
                
                    def content(self, content: Element) -> None:
                        # Custom content
                        Element("h1", content="Custom Content").classes("custom-content")
                """),
        ] = None,
        **kwargs,
    ):
        super().__init__(
            prefix=prefix,
            dependencies=dependencies,
            default_response_class=default_response_class,
            redirect_slashes=redirect_slashes,
            default=default,
            route_class=route_class,
            lifespan=lifespan,
            **kwargs,
        )
        if page_definition_class is not None and not issubclass(page_definition_class, PageDefinition):
            raise TypeError("page_definition_class must be a subclass of PageDefinition")
        self.page_definition_class = page_definition_class

    def page(
        self,
        path: str,
        *args,
        title: Optional[str] = None,
        page_definition_class: Optional[Type[PageDefinition]] = None,
        favicon: Optional[str] = None,
        router: Optional[APIRouter] = None,
        **kwargs,
    ) -> Callable:
        def decorator(func: Callable, *args, **kwargs) -> Callable:
            parameters_of_decorated_func = list(inspect.signature(func).parameters.keys())
            cap_title = title
            cap_page_definition_class = page_definition_class

            @functools.wraps(func)
            async def decorated(*dec_args, **dec_kwargs: DecKwargs) -> Response:
                Frame.get_stack().del_stack()
                # Create frame before function is called

                request = dec_kwargs["request"]
                response = dec_kwargs["response"]

                if self.page_definition_class is None:
                    self.page_definition_class = request.app.page_definition_class

                page_class = cap_page_definition_class or self.page_definition_class

                page = page_class()

                dec_kwargs = {k: v for k, v in dec_kwargs.items() if k in parameters_of_decorated_func}
                user_method = partial(func, *dec_args, **dec_kwargs)
                result = await page.render(user_method=user_method, request=request, title=cap_title)
                if isinstance(result, Response):
                    return self.return_function_response(result)
                standard_headers = {"cache-control": "no-store", "x-uiwiz-content": "page"}

                if response:
                    standard_headers.update(response.headers)

                self.add_ext(page, include_js=True, include_css=True)

                return HTMLResponse(
                    content=Frame.get_stack().render(),
                    status_code=200,
                    media_type="text/html",
                )

            self.__ensure_request_response_signature__(decorated)

            _router = router or self

            return _router.get(path, *args, include_in_schema=False, **kwargs)(decorated)

        return decorator

    def ui(
        self,
        path: str,
        include_js: bool = False,
        include_css: bool = False,
        router: Optional[APIRouter] = None,
        **kwargs,
    ) -> Callable:
        def decorator(func: Callable) -> Callable:
            # Capture values at decoration time
            cap_include_js = include_js
            cap_include_css = include_css
            parameters_of_decorated_func = list(inspect.signature(func).parameters.keys())

            @functools.wraps(func)
            async def decorated(*dec_args, **dec_kwargs) -> Response:
                Frame.get_stack().del_stack()
                Frame.get_stack()  # Create frame before function is called
                response = dec_kwargs["response"]

                # Ensure the signature matches the parameters of the function
                dec_kwargs = {k: v for k, v in dec_kwargs.items() if k in parameters_of_decorated_func}
                result = func(*dec_args, **dec_kwargs)
                if inspect.isawaitable(result):
                    result = await result

                if isinstance(result, Response):
                    Frame.get_stack().del_stack()
                    return result

                standard_headers = {"cache-control": "no-store", "x-uiwiz-content": "partial-ui"}
                standard_headers.update(response.headers)

                self.add_ext(page=None, include_js=cap_include_js, include_css=cap_include_css)

                content = Frame.get_stack().render()
                return HTMLResponse(content=content, headers=standard_headers)

            self.__ensure_request_response_signature__(decorated)
            _router = router or self
            return _router.post(path, include_in_schema=False, **kwargs)(decorated)

        return decorator

    def __ensure_request_response_signature__(self, func: Callable):
        data = {"request": Request, "response": Response}

        params = [p for p in inspect.signature(func).parameters.values()]
        for key, value in data.items():
            if key not in {p.name for p in params}:
                parm = inspect.Parameter(key, inspect.Parameter.POSITIONAL_OR_KEYWORD, annotation=value)
                params.insert(0, parm)

        func.__signature__ = inspect.Signature(params)

    def add_ext(
        self,
        page: Optional["PageDefinition"] = None,
        include_js: bool = False,
        include_css: bool = False,
    ) -> None:
        lib_css = []
        for lib in Frame.get_stack().extensions:
            if lib.endswith("css") and include_css:
                lib_css.append(lib)
                if page is None:
                    Element("link", href=lib, rel="stylesheet", type="text/css")
                else:
                    with page.header_ele:
                        Element("link", href=lib, rel="stylesheet", type="text/css")

                    # Hack to make aggrid work with daisyui
                    with page.html_ele:
                        Element("link", href=lib, rel="stylesheet", type="text/css")
            elif lib.endswith("js") and include_js:
                if page is None:
                    Element("script", src=lib, type="module")
                else:
                    with page.html_ele:
                        Element("script", src=lib, type="module")
            if not lib.endswith("js") and not lib.endswith("css"):
                raise Exception("lib type not supported, supported types css, js")



import importlib.metadata

__version__: str = "0.0.0"
try:
    # Could fail during development
    __version__ = importlib.metadata.version("uiwiz").replace("+", ".")
except importlib.metadata.PackageNotFoundError:
    pass



from typing import Optional

from uiwiz.element import Element
from uiwiz.elements.extensions.bindable import Bindable


class Label(Bindable):
    root_class: str = "label "
    root_size: str = "label-{size}"

    def __init__(self, text: Optional[str] = None, for_: Optional[Element] = None) -> None:
        """Label

        This element is used for labels that can be bound to form elements.

        .. code-block:: python
            from uiwiz import ui

            with ui.element().classes("flex flex-row gap-4"):
                label = ui.label("Username")
                input = ui.input("username")
                label.set_for(input)

            with ui.col():
                ui.label().combined_label("Password", ui.input("Password"))
                ui.label().combined_label("Type", ui.dropdown("dropdown", ["Option 1", "Option 2", "Option 3"], "Select an option"))
                ui.label().combined_label_end("Age", ui.input("Age"))
                

        :param text: The text to display
        :type text: str, optional
        :param for_: The element to bind the label to
        :type for_: Element, optional
        """
        super().__init__(tag="label")
        if text:
            self.content = text
        if for_:
            self._for = for_
            self.attributes["for"] = for_.id

        # Allign the label next to the element
        self.attributes["style"] = "padding-top: unset;"

    def set_text(self, text: str) -> "Label":
        self.content = text
        return self

    def set_for(self, for_: Element) -> "Label":
        self._for = for_
        self.attributes["for"] = for_.id
        return self
    
    def _combined_label(self, text: str, for_: Element, label_first: bool = True) -> "Label":
        """Helper to create a label with text and bind it to an element, order controlled by label_first."""
        if text:
            self.content = text
        # Remove from previous parents
        if self.parent_element and self in self.parent_element.children:
            self.parent_element.children.remove(self)
        if for_.parent_element and for_ in for_.parent_element.children:
            for_.parent_element.children.remove(for_)
        # Create container and add children in specified order
        with Element("label").classes(for_.tag) as container:
            if label_first:
                container.children.append(self)
                container.children.append(for_)
            else:
                container.children.append(for_)
                container.children.append(self)
            self.attributes["class"] = "label"
            for_.attributes["class"] = ""
        return self

    def combined_label(self, text: str, for_: Element) -> "Label":
        """Create a label with text and bind it to an element (label first)."""
        return self._combined_label(text, for_, label_first=True)

    def combined_label_end(self, text: str, for_: Element) -> "Label":
        """Create a label with text and bind it to an element (element first)."""
        return self._combined_label(text, for_, label_first=False)


from datetime import date, datetime
from typing import Optional, Union

from uiwiz.elements.extensions.on_event import OnEvent


class Datepicker(OnEvent):
    root_class: str = "input "

    def __init__(
        self,
        name: str,
        value: Optional[datetime] = None,
    ) -> None:
        """Datepicker element

        This element is used for date inputs

        .. code-block:: python
            from uiwiz import ui
            from datetime import datetime

            ui.datepicker("datepicker", datetime.now())

        :param name: name of the datepicker
        :type name: str
        :param value: default value of the datepicker
        :type value: datetime, optional
        """
        super().__init__("input")
        self.attributes["name"] = name
        self.attributes["type"] = "date"
        if value:
            self.default_date(value)

    def default_date(self, date: Union[datetime, date]) -> "Datepicker":
        if date is None:
            raise ValueError("Date cannot be None")
        self.value = date.strftime("%Y-%m-%d")
        return self



import pytest

from uiwiz import ui


def test_dict():
    dic = {"a": 1, "b": 2}
    output = str(ui.dict(dic))
    assert output is not None


def test_dict_exception():
    with pytest.raises(ValueError):
        ui.dict(None)


class Test:
    asd: int


def test_dict_exception_wrong_type():
    with pytest.raises(ValueError):
        ui.dict(Test())



from pathlib import Path
from typing import Callable, Optional, Union

from typing_extensions import override

from uiwiz import PageDefinition, ui
from uiwiz.svg.svg_handler import get_svg

parent = Path(__file__).parent
pages = []


class Page:
    def __init__(self, path: str, title: str, file: Union[Path, str, Callable[[], str], None] = None) -> None:
        self.path = path
        self.title = title
        if isinstance(file, Path):
            with file.open("r") as f:
                self.content = f.read()
        elif isinstance(file, str):
            self.content = file
        elif callable(file):
            self.content = file
        else:
            self.content = None

        if path not in [page.path for page in pages]:
            pages.append(self)

    async def render(self):
        """Render the page content."""
        with ui.container(padding="p-4"):
            if callable(self.content):
                await self.content()
            else:
                ui.markdown(self.content)


Page(path="/", title="Home", file=parent / "pages/index.md")


class Layout(PageDefinition):
    def __init__(self) -> None:
        """Layout

        A layout element that provides a consistent structure for the application.
        """
        super().__init__()
        self.drawer = None
        self._nav = None
        self.hide_on = "md"

    @override
    def content(self, _: ui.element) -> Optional[ui.element]:
        self.drawer = ui.drawer()
        with self.drawer:
            with self.drawer.drawer_content() as content:
                self.nav(self.drawer)

            with self.drawer.drawer_side():
                with ui.element("ul").classes("flex-none block md:hidden w-full"):
                    for page in pages:
                        with ui.element("li"):
                            ui.link(page.title, page.path)
                    with ui.element("li"):
                        ui.themeSelector()

        return content

    @override
    def footer(self, _: ui.element):
        with ui.footer().classes("footer mx-auto footer-center p-4 text-base-content"):
            with ui.element("div").classes("flex flex-col items-center justify-center"):
                ui.label("Made with ❤️ by Uiwiz").classes("text-sm")
                ui.link("GitHub", "https://github.com/declow/uiwizard")

    def nav(self, drawer):
        with ui.element().classes(
            "sticky top-0 flex h-16 justify-center bg-opacity-90 backdrop-blur transition-shadow duration-100 [transform:translate3d(0,0,0)] shadow-sm z-40"
        ):
            with ui.nav().classes("bg-base-200"):
                with ui.element().classes("flex-1"):
                    with ui.label(for_=drawer.drawer_toggle).classes(f"btn drawer-button {self.hide_on}:hidden"):
                        ui.html(get_svg("menu"))
                with ui.element().classes(f"flex-none hidden:{self.hide_on} block"):
                    with ui.element("ul").classes("menu menu-horizontal menu-md"):
                        for page in pages:
                            with ui.element("li"):
                                ui.link(page.title, page.path)
                        with ui.element("li"):
                            ui.themeSelector()



from bs4 import BeautifulSoup

from uiwiz import ui
from uiwiz.frame import Frame


def test_correct_parent():
    with ui.element() as root_1:
        with ui.element() as child_1_1:
            ui.element(content="test_1")

    with ui.element():
        with ui.element() as child_2_1:
            with child_1_1:
                ui.element(content="test_2")
            ui.element(content="test_3")

    soup = BeautifulSoup(Frame.get_stack().render(), "html.parser")

    assert len(child_2_1.children) == 1
    assert len(root_1.children) == 1
    assert (
        soup.prettify()
        == """<div id="a-0">
 <div id="a-1">
  <div id="a-2">
   test_1
  </div>
  <div id="a-5">
   test_2
  </div>
 </div>
</div>
<div id="a-3">
 <div id="a-4">
  <div id="a-6">
   test_3
  </div>
 </div>
</div>
"""
    )


def test_correct_parent_deep():
    with ui.element() as root_1:
        with ui.element() as child_1_1:
            ui.element(content="test_1")

    with ui.element():
        with ui.element() as child_2_1:
            with child_1_1:
                with ui.element(content="test_2"):
                    with ui.element() as child_2_2:
                        ui.element(content="test_4")
            ui.element(content="test_3")

    soup = BeautifulSoup(Frame.get_stack().render(), "html.parser")

    assert len(child_2_1.children) == 1
    assert len(root_1.children) == 1
    assert (
        soup.prettify()
        == """<div id="a-0">
 <div id="a-1">
  <div id="a-2">
   test_1
  </div>
  <div id="a-5">
   test_2
   <div id="a-6">
    <div id="a-7">
     test_4
    </div>
   </div>
  </div>
 </div>
</div>
<div id="a-3">
 <div id="a-4">
  <div id="a-8">
   test_3
  </div>
 </div>
</div>
"""
    )



import json
import logging
from mimetypes import guess_type
from pathlib import Path
from typing import Optional, Type, Union

from fastapi import FastAPI, Response
from fastapi.encoders import jsonable_encoder
from fastapi.exceptions import RequestValidationError
from fastapi.middleware.gzip import GZipMiddleware
from fastapi.responses import HTMLResponse
from fastapi.routing import APIRoute
from fastapi.staticfiles import StaticFiles
from starlette.requests import Request

from uiwiz.middleware.asgi_request_middleware import AsgiRequestMiddleware
from uiwiz.element import Element
from uiwiz.elements.button import Button
from uiwiz.elements.col import Col
from uiwiz.elements.html import Html
from uiwiz.frame import Frame
from uiwiz.page_route import PageDefinition, PageRouter
from uiwiz.shared import register_path, resources
from uiwiz.middleware.static_middleware import AsgiTtlMiddleware
from uiwiz.version import __version__

logger = logging.getLogger("uiwiz")
logger.addHandler(logging.NullHandler())


class CustomList(list):
    def append(self, route: APIRoute):
        if getattr(route, "endpoint", None):
            register_path(route.path, route.endpoint)
        return super().append(route)


class UiwizApp(FastAPI):
    def __init__(
        self,
        toast_delay: int = 2500,
        error_classes: str = "alert alert-error",
        cache_age: int = 14400,
        theme: Optional[str] = None,
        title: Optional[str] = "UiWiz",
        auto_close_toast_error: bool = False,
        page_definition_class: Type[PageDefinition] = PageDefinition,
        *args,
        **kwargs,
    ) -> None:
        """App used for asgi applications

        See fastapi documentation for more *args and **kwargs

        :param toast_delay: The time in milliseconds before the toast is removed
        :param error_classes: The classes to apply to the toast error for default validation errors
        :param cache_age: The time in seconds to cache the static files
        :param theme: The default theme to use
        :param title: The default title for the app
        :param auto_close_toast_error: If the toast error should auto close
        :param page_definition_class: The page definition class to use for the application
        :param args: FastAPI args
        :param kwargs: FastAPI kwargs
        """
        super().__init__(*args, **kwargs)
        self.router.routes = CustomList()
        self.toast_delay = toast_delay
        self.error_classes = error_classes
        self.auto_close_toast_error = auto_close_toast_error
        self.page_definition_class = page_definition_class
        self.theme = theme
        self.title = title
        self.add_static_files(f"/_static/{__version__}/", Path(__file__).parent / "static")

        self.add_middleware(AsgiRequestMiddleware)
        self.add_middleware(GZipMiddleware)
        self.add_middleware(AsgiTtlMiddleware, cache_age=cache_age)
        # self.add_middleware(StripHiddenFormFieldMiddleware)
        self.extensions: dict[str, Path] = {}
        self.app_paths: dict[str, Path] = {}

        self.exception_handler(RequestValidationError)(self.handle_validation_error)

        @self.get("/_static/extension/{__version__}/{extension}/{filename}", include_in_schema=False)
        def get_extension(extension: str, filename: str):
            resource_key = f"{extension}/{filename}"
            if resource_key not in resources:
                return Response(status_code=404)

            with open(resources[resource_key], encoding="utf-8") as f:
                content = f.read()

            content_type, _ = guess_type(resource_key)
            return Response(content, media_type=content_type)

    def add_static_files(self, url_path: str, local_directory: Union[str, Path]) -> None:
        self.mount(url_path, StaticFiles(directory=str(local_directory)))

    def page(
        self, path: str, *args, title: Optional[str] = None, favicon: Optional[str] = None, **kwargs
    ) -> PageRouter:
        return PageRouter(page_definition_class=self.page_definition_class).page(
            path, *args, title=title, favicon=favicon, router=self.router, **kwargs
        )

    def ui(self, path: str, *args, include_js: bool = True, include_css: bool = True, **kwargs) -> PageRouter:
        return PageRouter().ui(path=path, include_js=include_js, include_css=include_css, router=self.router, **kwargs)

    async def handle_validation_error(self, request: Request, exc: RequestValidationError):
        fields_with_errors = [item.get("loc")[1] for item in exc.errors()]
        ok_fields = [item for item in exc.body.keys() if item not in fields_with_errors]

        Frame.get_stack().del_stack()
        Frame.get_stack()

        with Element().classes(self.error_classes) as toast:
            toast.attributes["id"] = "toast"
            toast.attributes["hx-swap-oob"] = "afterbegin"
            toast.attributes["hx-toast-data"] = json.dumps(
                jsonable_encoder(
                    {
                        "detail": exc.errors(),
                        "fieldErrors": fields_with_errors,
                        "fieldOk": ok_fields,
                    }
                )
            )
            html = Html("").classes("alert alert-error relative")
            html.tag = "span"
            html.attributes["hx-toast-data"] = json.dumps({"autoClose": self.auto_close_toast_error})
            html.attributes["hx-toast-delete-button"] = lambda: btn.id
            with html:
                with Col(gap="").classes("relative"):
                    for item in exc.errors():
                        Element(content=f"{item.get('loc')[1]}: {item.get('msg')}")
                if not self.auto_close_toast_error:
                    btn = Button("✕").classes("btn btn-sm btn-circle btn-ghost absolute right-2 top-2")

        html_content = Frame.get_stack().render()

        return HTMLResponse(
            content=html_content,
            status_code=400,
            headers={
                "cache-control": "no-store",
                "x-uiwiz-content": "page",
                "x-uiwiz-validation-error": "true",
            },
        )



import html

from uiwiz import ui


def test_raw_html():
    html = ui.html("<strong>test</strong>")
    assert '<div id="a-0"><strong>test</strong></div>' == str(html)


def test_escaped_html():
    input = "<strong>test</strong>"
    _html = ui.element(content=input)
    assert f'<div id="a-0">{html.escape(input)}</div>' == str(_html)



from uiwiz.element import Element


class FullWidth(Element):
    root_class = "w-full"

    def __init__(self) -> None:
        """FullWidth

        This element is used to create a full-width container.

        .. code-block:: python
            from uiwiz import ui

            with ui.full_width():
                ui.text("This is a full-width container")

        """
        super().__init__()
        self.classes(FullWidth.root_class)



import re
import textwrap


def extract_text(docstring: str) -> tuple[str, str, dict[str, str]]:
    docstring = textwrap.dedent(docstring or "").strip()
    lines = docstring.splitlines()

    # Find code block (robust to any indentation, blank lines, and multiple code blocks)
    code_block = ""
    code_block_lines = []
    in_code_block = False
    code_block_indent = None
    for i, line in enumerate(lines):
        if not in_code_block and re.match(r"\..\s*code-block", line.strip(), re.IGNORECASE):
            # Find the first non-blank, indented line after the marker
            for j in range(i + 1, len(lines)):
                if lines[j].strip() == "":
                    continue
                indent = len(lines[j]) - len(lines[j].lstrip())
                if indent > 0:
                    code_block_indent = indent
                    in_code_block = True
                    start = j
                    break
            if in_code_block:
                # Collect all lines with at least this indent (or blank)
                for k in range(start, len(lines)):
                    line_k = lines[k]
                    if line_k.strip() == "":
                        code_block_lines.append("")
                        continue
                    if len(line_k) - len(line_k.lstrip()) >= code_block_indent:
                        code_block_lines.append(line_k[code_block_indent:])
                    else:
                        break
            break
    if code_block_lines:
        # Remove leading/trailing blank lines in code block
        while code_block_lines and code_block_lines[0].strip() == "":
            code_block_lines.pop(0)
        while code_block_lines and code_block_lines[-1].strip() == "":
            code_block_lines.pop()
        code_block = "\n".join(code_block_lines)

    # Extract parameter lines that start with ':param' (allow for type or not)
    parameters = {}
    for line in lines:
        m = re.match(r":param(?:\s+\w+)?\s+(\w+)\s*:\s*(.*)", line.strip())
        if m:
            param_name, param_desc = m.groups()
            parameters[param_name] = param_desc.strip()

    # Build the description by excluding the code block and param lines
    description_lines = []
    in_code_block = False
    for i, line in enumerate(lines):
        stripped = line.strip()
        if re.match(r"\..\s*code-block", stripped, re.IGNORECASE):
            in_code_block = True
            continue
        if in_code_block:
            if line.strip() == "":
                continue
            indent = len(line) - len(line.lstrip())
            if indent >= (code_block_indent or 1):
                continue
            else:
                in_code_block = False
        if re.match(r":param(?:\s+\w+)?\s+\w+\s*:", stripped):
            continue
        if not in_code_block:
            stripped = stripped.lstrip()
            if stripped:
                description_lines.append(stripped)
    description = "\n\n".join(description_lines).strip()

    return description, code_block, parameters



import pytest

from uiwiz.svg.svg_handler import get_svg


def test_get_svg():
    error = get_svg("error")
    svg_error = """<svg xmlns="http://www.w3.org/2000/svg" class="stroke-current shrink-0 h-6 w-6" fill="none"
    viewBox="0 0 24 24">
    <path stroke-linecap="round" stroke-linejoin="round" stroke-width="2"
        d="M10 14l2-2m0 0l2-2m-2 2l-2-2m2 2l2 2m7-2a9 9 0 11-18 0 9 9 0 0118 0z" />
</svg>"""
    assert svg_error == error


def test_get_svg_empty_string():
    with pytest.raises(ValueError):
        get_svg("")


def test_get_svg_none():
    with pytest.raises(ValueError):
        get_svg(None)



import pytest

from uiwiz.docs.pages.docs.extract_doc import extract_text
from uiwiz import ui


@pytest.mark.parametrize("docstring, expected_code_snippet", [
    (ui.ace.__init__.__doc__, 'ui.ace(name="editor")'),
    (ui.upload.__init__.__doc__, 'ui.upload("file")'),
    (ui.button.__init__.__doc__, 'button'),
    (ui.textarea.__init__.__doc__, 'ui.textarea'),
    (ui.input.__init__.__doc__, 'from uiwiz import ui\n\nui.input("username", "default_value", "Enter your username")'),
    (ui.upload.on_upload.__doc__, 'ui.upload("file").on_upload(on_upload=handle_upload, swap="none")'),
    #(ui.toast.__init__.__doc__, 'ui.toast("message")'),
    #(ui.aggrid.__init__.__doc__, 'ui.aggrid("grid")'),
])
def test_extract_text(docstring, expected_code_snippet):
    description, code_block, parameters = extract_text(docstring)

    # Verify that the description has no leading/trailing whitespace,
    # and does not include the code block or parameter definitions.
    assert "code-block" not in description
    assert ":param" not in description

    # If the docstring has a code block, check for the expected snippet
    if '.. code-block' in docstring:
        assert expected_code_snippet in code_block
        if 'from uiwiz import ui' in docstring:
            assert 'from uiwiz import ui' in code_block
    else:
        assert code_block == ""

    # Parameter checks only for Ace (others may not have all these params)
    if expected_code_snippet == 'ui.ace(name="editor")':
        expected_params = {
            "name": "The name of the textarea element to be submitted with a form",
            "content": "The initial content of the editor",
            "lang": 'The language mode to use. One of "sql" or "python"',
            "sql_options": "Options for the SQL language mode. Tables and columns to be used for autocompletion",
            "ace_options": "Options for the Ace Editor",
        }
        assert parameters == expected_params


from typing import Optional, Union

from uiwiz.element import Element
from uiwiz.elements.extensions.on_event import OnEvent
from collections import namedtuple

DropdownItem = namedtuple("DropdownItem", ["name", "value"])

class Dropdown(OnEvent):
    root_class: str = "select "
    root_size: str = "select-{size}"
    _classes: str = "select-bordered w-full max-w-xs"

    def __init__(self, name: str, items: Union[list[str], list[DropdownItem]], placeholder: Optional[str] = None) -> None:
        """Dropdown

        A dropdown is a list in which the selected item is always visible, and the others are visible on demand.

        .. code-block:: python
            from uiwiz import ui

            ui.dropdown("dropdown", ["Option 1", "Option 2", "Option 3"], "Select an option")
            ui.dropdown("dropdown", ["Option 1", "Option 2", "Option 3"], "Option 1") # Option 1 is selected by default
            ui.dropdown("dropdown", [ui.dropdownItem(name="Option 1", value="1"), ui.dropdownItem(name="Option 2", value="2")], "Select an option")

        :param name: The name of the dropdown
        :type name: str
        :param items: List of items in the dropdown
        :type items: Union[list[str], list[DropdownItem]]
        :param placeholder: Placeholder text
        :type placeholder: str, optional
        """
        super().__init__("select")
        self.attributes["name"] = name
        self.classes()
        self.size(self._size)
        with self:
            if placeholder and placeholder not in items:
                Element("option disabled selected", content=placeholder)
            for v in items:
                if isinstance(v, DropdownItem):
                    e = Element("option", content=v.name)
                    e.value = v.value
                    if placeholder == v.value:
                        e.attributes["selected"] = "selected"
                else:
                    e = Element("option", content=v)
                    e.value = v
                    if placeholder == v:
                        e.attributes["selected"] = "selected"



from fastapi.testclient import TestClient

from uiwiz import ui
from uiwiz.app import UiwizApp
from uiwiz.page_route import PageRouter
from uiwiz.shared import fetch_route


def test_page_go_to_get_request():
    pr = PageRouter()
    route = "/path"

    @pr.page(route)
    def func(): ...  # pragma: no cover

    app = UiwizApp()
    app.include_router(pr)

    apis = {item.path: item for item in app.routes}

    assert route == fetch_route(func)
    assert route == apis.get(route).path
    assert "func" == apis.get(route).name
    assert {"GET"} == apis.get(route).methods


def test_ui_go_to_post_request():
    app = UiwizApp()

    pr = PageRouter()
    route = "/path"

    @pr.ui(route)
    def func(): ...  # pragma: no cover

    app.include_router(pr)

    apis = {item.path: item for item in app.routes}

    assert route == fetch_route(func)
    assert route == apis.get(route).path
    assert "func" == apis.get(route).name
    assert {"POST"} == apis.get(route).methods


def test_app_ui_go_to_post_request():
    app = UiwizApp()

    route = "/path"

    @app.ui(route)
    def func(): ...  # pragma: no cover

    apis = {item.path: item for item in app.routes}

    assert route == fetch_route(func)
    assert route == apis.get(route).path
    assert "func" == apis.get(route).name
    assert {"POST"} == apis.get(route).methods


def test_router_ui_prefix():
    app = UiwizApp()

    pr = PageRouter(prefix="/api")
    route = "/path"

    @pr.ui(route)
    def func(): ...  # pragma: no cover

    app.include_router(pr)
    apis = {item.path: item for item in app.routes}
    full_route = pr.prefix + route

    assert full_route == fetch_route(func)
    assert full_route == apis.get(full_route).path
    assert "func" == apis.get(full_route).name
    assert {"POST"} == apis.get(full_route).methods


def test_router_page_prefix():
    app = UiwizApp()

    pr = PageRouter(prefix="/page")
    route = "/path"

    @pr.page(route)
    def func(): ...  # pragma: no cover

    app.include_router(pr)
    apis = {item.path: item for item in app.routes}
    full_route = pr.prefix + route

    assert full_route == fetch_route(func)
    assert full_route == apis.get(full_route).path
    assert "func" == apis.get(full_route).name
    assert {"GET"} == apis.get(full_route).methods


def test_router_page_without_prefix():
    app = UiwizApp()

    pr = PageRouter()
    route = "/path"

    @pr.page(route)
    def func(): ...  # pragma: no cover

    app.include_router(pr)
    apis = {item.path: item for item in app.routes}
    full_route = pr.prefix + route

    assert full_route == fetch_route(func)
    assert full_route == apis.get(full_route).path
    assert "func" == apis.get(full_route).name
    assert {"GET"} == apis.get(full_route).methods


def test_router_ui_without_extensions():
    app = UiwizApp()

    pr = PageRouter()
    route = "/path"

    @pr.ui(route, include_js=False, include_css=False)
    def func():
        ui.markdown("""
# Test
## Test
### Test
""")

    app.include_router(pr)

    client = TestClient(app)
    response = client.post(route)
    body = response.read().decode("utf-8")
    assert response.status_code == 200
    assert "Markdown/markdown.css" not in body
    assert "Markdown/codehighlight.css" not in body


def test_router_ui_extensions():
    app = UiwizApp()

    pr = PageRouter()
    route = "/path"

    @pr.ui(route)
    def func():
        ui.markdown("""
# Test
## Test
### Test
""")

    app.include_router(pr)

    client = TestClient(app)
    response = client.post(route)
    body = response.read().decode("utf-8")
    assert response.status_code == 200
    assert "Markdown/markdown.css" not in body
    assert "Markdown/codehighlight.css" not in body



import inspect
import json
from html import escape
from typing import Annotated, Callable, Optional

from fastapi import Depends, Request, Response

from uiwiz.element import Element
from uiwiz.frame import Frame
from uiwiz.version import __version__


class PageDefinition:
    html_ele: Element
    header_ele: Element
    body_ele: Element
    content_ele: Element
    title_ele: Element
    lang: str

    def __init__(self):
        """
        The PageDefinition class is designed to be subclassed, allowing
        developers to override the `header`, `body`, and `content` methods
        to customize the HTML structure and content as needed. The `footer`
        method can also be overridden to add custom footer content.

        This allows for a flexible and reusable page structure that can
        be easily extended for different pages in a web application or applied
        to all pages in a web application by setting the `page_definition_class`
        in the `UiwizApp` instance.


        ## Example:
        ```python
        from uiwiz import PageDefinition, Element

        class MyPage(PageDefinition):
            def header(self, header: Element) -> None:
                Element("link", href="/custom.css", rel="stylesheet")

            def body(self, body: Element) -> None:
                Element("div", content="Custom Body").classes("custom-body")

            def content(self, content: Element) -> Optional[Element]:
                return Element("h1", content="Custom Content").classes("custom-content")

            def footer(self, content: Element) -> None:
                Element("footer", content="Custom Footer").classes("custom-footer")
        ```

        """
        self._lang: str = "en"

    @property
    def lang(self) -> str:
        return self._lang

    @lang.setter
    def lang(self, value: str) -> None:
        self._lang = value
        self.html_ele.attributes["lang"] = value

    @property
    def title(self) -> str:
        return self.title_ele.content

    @title.setter
    def title(self, value: str) -> None:
        self.title_ele.content = value

    async def render(
        self,
        user_method: Optional[Callable],
        request: Request,
        title: Optional[str] = None,
    ) -> Optional[Response]:
        frame = Frame.get_stack()

        theme = request.app.theme
        if cookie_theme := request.cookies.get("data-theme"):
            theme = escape(cookie_theme)

        class RenderDoctype:
            def render(self):
                return "<!DOCTYPE html>"

        frame.root.append(RenderDoctype())  # funky way to add doctype
        page_title = request.app.title if title is None else title
        with Element("html").classes("overflow-y-scroll") as html:
            self.html_ele = html

            html.attributes["id"] = "html"
            html.attributes["lang"] = self.lang
            if theme:
                html.attributes["data-theme"] = theme

            with Element("head") as header:
                self.header_ele = header

                Element("meta", name="viewport").attributes["content"] = "width=device-width, initial-scale=1"
                Element("meta", charset="utf-8")
                Element("meta", description=frame.meta_description_content)

                self.title_ele = Element("title", content=page_title)

                Element("link", href=f"/_static/{__version__}/libs/output.css", rel="stylesheet", type="text/css")
                Element("link", href=f"/_static/{__version__}/libs/daisyui.css", rel="stylesheet", type="text/css")
                Element(
                    "link", href=f"/_static/{__version__}/libs/daisyui-themes.css", rel="stylesheet", type="text/css"
                )
                Element("script", src=f"/_static/{__version__}/libs/tailwind.js")
                Element("link", href=f"/_static/{__version__}/app.css", rel="stylesheet", type="text/css")
                self.header(header)
            with Element("body") as body:
                self.body_ele = body

                body.attributes["hx-ext"] = "swap-header"
                self.body(body)
                with Element("div").classes("flex flex-col w-full min-h-screen") as content:
                    self.content_ele = content
                    user_content = self.content(content)
                    if user_content is not None:
                        if isinstance(user_content, Element):
                            self.content_ele = user_content
                        else:
                            raise TypeError(f"Expected Element, got {type(user_content).__name__} in content method")
                    with self.content_ele:
                        result = user_method()
                        if inspect.isawaitable(result):
                            result = await result
                        self.footer(content)

                toast = Element("div").classes("toast toast-top toast-end text-wrap z-50")
                toast.attributes["id"] = "toast"
                toast.attributes["hx-toast-delay"] = json.dumps({"delay": request.app.toast_delay})

            Element("script", src=f"/_static/{__version__}/libs/htmx2.0.7.min.js")
            Element("script", src=f"/_static/{__version__}/libs/htmx-json-enc.js")
            Element("script", src=f"/_static/{__version__}/default.js")

        return result

    def header(self, header: Element) -> None:
        pass

    def body(self, body: Element) -> None:
        pass

    def content(self, content: Element) -> Optional[Element]:
        pass

    def footer(self, content: Element) -> None:
        pass


Page = Annotated[PageDefinition, Depends()]



from typing import Optional

from uiwiz.element import Element
from uiwiz.event import FUNC_TYPE, SWAP_EVENTS, TARGET_TYPE


class Form(Element):
    root_class: str = "flex flex-col items-start gap-4 p-4 "

    def __init__(self) -> None:
        """Form

        A form is a section of a document containing input elements.

        .. code-block:: python
            from uiwiz import ui

            @app.ui("/ui/form/submit")
            async def ui_form_submit():
                ui.toast("Form submitted successfully!").success()

            with ui.form().on_submit(ui_form_submit):
                ui.input(name="name", placeholder="Name")
                ui.button("Submit")
        """
        super().__init__("form")

    def on_submit(
        self,
        func: FUNC_TYPE = None,
        target: TARGET_TYPE = None,
        swap: SWAP_EVENTS = "none",
        params: Optional[dict[str, str]] = None,
    ):
        self.event = {"func": func, "trigger": "submit", "target": target, "swap": swap, "params": params}
        return self



from uiwiz.elements.extensions.on_event import OnEvent


class Checkbox(OnEvent):
    root_class: str = "checkbox"
    root_size: str = "checkbox-{size}"

    def __init__(self, name: str, checked: bool = False) -> None:
        """Checkbox element

        Used for forms where a user can select one or more options.
        .. code-block:: python
            from uiwiz import ui

            ui.checkbox("subscribe", checked=True)

        :param name: name of the checkbox
        :type name: str
        :param checked: checked status
        :type checked: bool, optional
        """
        super().__init__("input")
        if checked:
            self.attributes["checked"] = "checked"
        self.attributes["name"] = name
        self.attributes["type"] = "checkbox"



import html
import json
from enum import Enum
from pathlib import Path
from typing import Any, Tuple

import numpy as np
import pandas as pd
from fastapi.responses import JSONResponse

from uiwiz.element import Element

LIB_PATH = Path(__file__).parent / "aggrid-community.min.js"
CSS_PATH = Path(__file__).parent / "aggridtheme.css"
JS_PATH = Path(__file__).parent / "aggrid.js"

class OPTIONS(str, Enum):
    autoSizeColumn = "autoSizeAll"
    fitColumnContent = "sizeToFit"

class Aggrid(Element, extensions=[CSS_PATH, LIB_PATH, JS_PATH]):


    _classes: str = "ag-theme-quartz ag-theme-uiwiz w-full"

    def __init__(self, df: pd.DataFrame) -> None:
        """Aggrid

        Use aggrid to display a DataFrame in a grid format.
        Can be used anywhere in a @page_router.ui("<path>") or @app.ui("<path>")

        Example:
        .. code-block:: python
        
            from uiwiz import ui
            import pandas as pd

            df = pd.DataFrame({
                "Name": ["Alice", "Bob", "Charlie"],
                "Age": [25, 30, 35],
                "City": ["New York", "Los Angeles", "Chicago"]
            })
            ui.aggrid(df) 

        :param df: The DataFrame to display in the grid
        """
        super().__init__("div")
        self.classes(Aggrid._classes)

        cols, rows = Aggrid.create_cols_and_rows(df)

        self.attributes["hx-ext"] = "hx-aggrid"
        self.attributes.__setitem__("hx-aggrid-cols", cols, False)
        self.attributes.__setitem__("hx-aggrid-rows", rows, False)
        self.attributes["hx-aggrid"] = "/data"

    @staticmethod
    def create_cols_and_rows(df: pd.DataFrame, escape: bool = True) -> Tuple[list[Any], list[Any]]:
        cols = []
        rows = []

        if df is not None:
            df = df.replace({np.nan: None})
            cols = [{"field": item} for item in df.columns.to_list()]
            rows = df.to_dict("records")
            if escape:
                cols = html.escape(json.dumps(cols))
                rows = html.escape(json.dumps(rows, default=str))

        return cols, rows

    @staticmethod
    def response(df: pd.DataFrame, headers: dict[str, str] = {}) -> JSONResponse:
        _headers = {"HX-Trigger": "aggridUpdate"} | headers
        cols, rows = Aggrid.create_cols_and_rows(df, False)
        return JSONResponse({"cols": cols, "rows": rows}, headers=_headers)



from typing import Callable, Union

from uiwiz.element import Element


class Link(Element):
    root_class: str = "link "
    _classes: str = "link-hover"

    def __init__(self, text: str, link: Union[Callable[[], str], str]) -> None:
        """Link element

        Link element that can be used to navigate to a different page.
        Can be used as a button or a link.

        Example:
        .. code-block:: python
        
            from uiwiz import ui

            ui.link("Click me", "/some/path").classes("btn btn-primary") 

        :param text: The text to display
        :type text: str
        :param link: The link to navigate to when clicked. Can be a callable that returns a string
        :type link: Union[Callable[[], str], str]
        """
        super().__init__("a")
        self.content = text
        self.link = link
        if isinstance(link, str):
            self.attributes["href"] = link
        self.classes(Link._classes)

    def before_render(self) -> None:
        if callable(self.link):
            self.attributes["href"] = self.link()
            return
        self.attributes["href"] = self.link



import inspect
import typing
from pathlib import Path

from docs.pages.docs.extract_doc import extract_text
from uiwiz import PageRouter, elements, ui


def get_class_properties(cls):
    # Returns a list of (name, property) tuples for all properties in the class
    return [
        (name, prop)
        for name, prop in inspect.getmembers(cls)
        if not name.startswith("_")
        and not inspect.isroutine(prop)
        and not isinstance(prop, (property, staticmethod, classmethod))
    ]

def get_clean_annotation_name(annotation):
    import typing
    if hasattr(annotation, "__name__"):
        return annotation.__name__
    elif hasattr(annotation, "_name") and annotation._name:
        return annotation._name
    elif hasattr(typing, "ForwardRef") and isinstance(annotation, typing.ForwardRef):
        return annotation.__forward_arg__
    else:
        return str(annotation).replace("typing.", "")


def extract_param_annotations(cls):
    sig = inspect.signature(cls.__init__)
    annotations = {}
    for name, param in sig.parameters.items():
        if param.annotation is not inspect.Parameter.empty:
            default_value = str(param.default)
            annotations[name] = {
                "type": get_clean_annotation_name(param.annotation),
            }
            if param.default is not inspect.Parameter.empty:
                annotations[name]["default"] = default_value
    return annotations


def create_elements(router: PageRouter):
    print(Path(elements.__file__).parent)
    for element_name in dir(ui):
        if element_name.startswith("_"):
            continue
        element = getattr(ui, element_name)
        create_docs_element(element, router)


def create_docs_element(element: ui.element, router: PageRouter):
    app = router # noqa
    with ui.container(space_y="").classes("prose rounded-lg"):
        with ui.element().classes("flex flex-row"):
            ui.element("h2", f"ui.{element.__name__.lower()}")

        des, cb, _ = extract_text(element.__init__.__doc__)
        ui.markdown(des).classes("text-content")
        with ui.element().classes("not-prose"):
            try:
                ui.markdown("""```python
""" + cb + " " + """```""")
            except Exception:
                pass
            try:
                exec(cb)
            except Exception:
                pass

        ui.element("h3", "Constructor").classes("mt-4")
        anno = extract_param_annotations(element)
        if anno:
            for name, details in anno.items():
                with ui.element("div").classes("flex flex-row pl-4"):
                    ui.element("span", f"{name}: {details['type']}").classes("font-bold")
                    if "default" in details:
                        ui.element("span", f"= {details['default']}").classes("text-gray-500 ml-2")
                    else:
                        ui.element("span", "No default required argument").classes("text-gray-500 ml-2")
            
            methods = [method for method in inspect.getmembers(element, predicate=inspect.isfunction) if not method[0].startswith("_")]

            if methods:
                with ui.element("h3").classes("mt-4"):
                    ui.element("span", "Methods")
                for method_name, method in methods:
                    with ui.element("div").classes("flex flex-row"):
                        sig = inspect.signature(method)
                        # Use get_type_hints to resolve forward references
                        try:
                            type_hints = typing.get_type_hints(method, globalns=method.__globals__, localns=vars(element))
                        except Exception:
                            type_hints = {}
                        params = []
                        for name, param in sig.parameters.items():
                            if name == "self":
                                continue
                            annotation = type_hints.get(name, param.annotation)
                            param_type = get_clean_annotation_name(annotation) if annotation is not inspect.Parameter.empty else 'Any'
                            params.append(f"{name}: {param_type}")
                        param_str = ", ".join(params)
                        # Return type
                        ret_anno = type_hints.get('return', sig.return_annotation)
                        return_type = get_clean_annotation_name(ret_anno) if ret_anno is not inspect.Signature.empty else "Any"
                        # Method name in bold, params in blue, return type in green
                        ui.element("span", f"{method_name}(").classes("font-bold pl-4")
                        if param_str:
                            ui.element("span", param_str).classes("text-info ml-1 mr-1")
                        ui.element("span", ")").classes("font-bold")
                        ui.element("span", f" -> {return_type}").classes("text-success ml-2")
                    if not method.__doc__:
                        ui.element("span", content="No documentation provided").classes("text-gray-500")
                    else:
                        for line in method.__doc__.splitlines():
                            ui.element("div", content=line).classes("pl-8")



from uiwiz.element import Element


class Nav(Element):
    root_class: str = "navbar w-full gap-4"
    root_size: str = "navbar-{size}"

    def __init__(self) -> None:
        """Nav
        
        This element is used for navigation bars

        .. code-block:: python
            from uiwiz import ui

            with ui.nav():
                ui.link("Home", "/")
                ui.link("Docs", "/docs")
        
        """
        super().__init__()



from starlette.middleware.base import BaseHTTPMiddleware
from starlette.requests import Request
from starlette.types import ASGIApp
import json

class StripHiddenFormFieldMiddleware(BaseHTTPMiddleware):
    def __init__(self, app: ASGIApp, field_name: str = "csrf_token"):
        super().__init__(app)
        self.field_name = field_name

    async def dispatch(self, request: Request, call_next):
        if request.method in ("POST", "PUT", "PATCH") and request.headers.get("content-type", "").startswith("application/json"):
            body: dict = await request.json()
            body.pop(self.field_name, None)
            new_body = json.dumps(body).encode()

            async def receive():
                return {"type": "http.request", "body": new_body, "more_body": False}
            request._receive = receive

        response = await call_next(request)
        return response


from uiwiz.element import Element


class Html(Element):
    def __init__(self, content) -> None:
        """Html element

        Will render a raw htlm string. This is useful for embedding custom HTML content directly into the UI.

        .. code-block:: python
            from uiwiz import ui

            ui.html("<strong>Hello World</strong>")

        :param content: html content as string
        :type content: str
        """
        super().__init__()
        self.__content__ = content

    @property
    def content(self) -> str:
        return self.__content__

    @content.setter
    def content(self, content):
        self.__content__ = content



from typing import Any, Optional

from uiwiz.element import Element


class HiddenInput(Element):
    _classes: str = ""

    def __init__(
        self,
        name: str,
        value: Optional[Any] = None,
    ) -> None:
        """HiddenInput

        This element is used for hidden input data that should not be visible to the user.
        But it will be sent back to the server when the form is submitted.

        Example:
        .. code-block:: python

            from uiwiz import ui
            from pydantic import BaseModel

            class HiddenInputExample(BaseModel):
                csrf_token: str
                age: int
                

            @app.ui("/ui/hiddeninput/submit")
            async def ui_hiddeninput_submit(data_input: HiddenInputExample):
                with ui.toast().set_auto_close(False).success():
                    ui.dict(data_input.dict()).border_classes("")

            with ui.form().on_submit(ui_hiddeninput_submit):
                ui.hiddenInput(name="csrf_token", value="hidden_csrf_token")
                ui.input("age", None, "Enter your age")
                ui.button("Submit")

        :param name: Used in the attributes name sent back in json
        :type name: str
        :param value: The value to send back
        :type value: Any
        """
        super().__init__("input")
        self.attributes["name"] = name
        self.attributes["value"] = value
        self.attributes["type"] = "hidden"



from pathlib import Path

from docs.layout import Layout, Page
from docs.pages.docs.elements import create_docs_element, create_elements
from uiwiz import PageRouter, ui

parent = Path(__file__).parent


class LayoutDocs(Layout):
    def after_render(self, request):
        self.drawer.always_open(True)


docs_router = PageRouter(prefix="/reference")


@docs_router.page("/element/{name}", title="Elements")
async def docs_element(name: str):
    create_docs_element(getattr(ui, name), docs_router)


@docs_router.page("/elements", title="Elements")
async def docs_elements():
    create_elements(docs_router)


async def reference():
    with ui.container(padding="p-4"):
        ui.markdown("This is the reference page. It will contain links to all the elements and their documentation.")
        for value in dir(ui):
            if value.startswith("_"):
                continue
            ui.link(value, f"{docs_router.prefix}/element/{value}")
        ui.link("All Elements", f"{docs_router.prefix}/elements")


Page(path=f"{docs_router.prefix}/", title="Reference", file=reference)
